import re

from docutils import nodes
from docutils.nodes import Node, Element
from docutils.parsers.rst import directives, Directive
from sphinx import addnodes
from sphinx.addnodes import desc_signature, pending_xref, desc_content
from sphinx.application import Sphinx
from sphinx.builders import Builder
from sphinx.directives import ObjectDescription, ObjDescT
from sphinx.domains import Domain, ObjType
from sphinx.environment import BuildEnvironment
from sphinx.roles import XRefRole
from sphinx.util.docfields import Field, TypedField, DocFieldTransformer
from sphinx.util.nodes import make_refnode

from .utils import to_safe_label


class VBXRefRole(XRefRole):
    '''For VBDomain's cross-reference e.g. func role (vb:func).
    '''
    # Visual Basic 固有のリンク処理が必要な場合は process_link() を
    # オーバーライドする。
    # TODO: 不要ならこのクラスを削除する。
    pass


class VBFunction(ObjectDescription):
    '''For function directive (vb:function).
    '''
    # Class variables are function directive's specs.
    has_content = True  # for function summary or details.
    required_arguments = 1  # 1 arg is enough as whole declaration.
    option_spec = {
        'module': directives.unchanged,
    }

    doc_field_types = [
        TypedField(
            'parameter', label='パラメータ',
            names=('param', 'parameter', 'arg', 'argument'),
            typerolename='vbtype', typenames=('type',)),
        Field('returnvalue', label='戻り値', has_arg=False,
              names=('returns', 'return')),
        Field('returntype', label='戻り値の型', has_arg=False,
              names=('rtype',)),
    ]

    def handle_signature(
            self, sig: str, signode: desc_signature) -> ObjDescT:
        '''Parse the function signature and add nodes to the signode.

        This function will be called from autodoc.

        Parameters
        ----------
        sig : str
            Function declaration from VB source code.
        signode : desc_signature
            Node for signature structure.

        Returns
        -------
        obj_id : ObjDescT
            Object identifier.
        '''
        # Split signature into 3 parts.
        match = re.match(r'^(.*?)\s*(\(.*?\))?\s*(As\s+.+)?$', sig)
        if not match:
            raise ValueError(f"Invalid signature: {sig}")

        # Part 1: access modifier and function name.
        access_and_name = match.group(1).strip()
        # Part 2: parameters (without ()).
        parameters = match.group(2).strip("()") if match.group(2) else ""
        if parameters:
            args = [arg.strip() for arg in parameters.split(',')]
        else:
            args = []
        # Part 3: return type.
        return_type = match.group(3).removeprefix("As").strip() \
            if match.group(3) else ""

        # Extract access modifier.
        valid_modifiers = [
            'Public', 'Private', 'Protected', 'Friend', 'Protected Friend',
            'Private Protected']

        parts = access_and_name.split()

        if len(parts) >= 2 and ' '.join(parts[:2]) in valid_modifiers:
            access_modifier = ' '.join(parts[:2])
            parts = parts[2:]
        elif parts[0] in valid_modifiers:
            access_modifier = parts.pop(0)
        else:
            access_modifier = ''

        # Extract function type ("Function" or "Sub").
        func_type = parts.pop(0) if len(parts) > 0 else ''
        if func_type not in ('Function', 'Sub'):
            raise ValueError(f'No "Function" or "Sub" in "{sig}".')

        # Extract function name.
        func_name = parts.pop(0) if len(parts) > 0 else ''
        if not func_name:
            raise ValueError(f'No Function name in "{sig}".')

        # Add access modifier to signode (e.g. 'Public ').
        if access_modifier:
            signode += addnodes.desc_annotation(
                access_modifier + ' ', access_modifier + ' ')

        # Add function type to signode (e.g. 'Function ').
        signode += addnodes.desc_annotation(func_type + ' ', func_type + ' ')

        # Add function name to signode (e.g. 'MyFunc').
        signode += addnodes.desc_name(func_name, func_name)

        # Add param list to signode.
        paramlist = addnodes.desc_parameterlist()
        for arg in args:
            # Add param (e.g. 'ByVal arg1 As Integer').
            param = addnodes.desc_parameter('', arg)
            paramlist += param
        signode += paramlist

        # Add return type to signode (e.g. ' As String').
        if return_type:
            signode += nodes.emphasis(
                ' As ' + return_type, ' As ' + return_type)

        # Return Object identifier.
        module_name = self.options.get('module', '')
        obj_id = f'{module_name}.{func_name}' if module_name else func_name
        return obj_id

    def transform_content(self, contentnode: desc_content):
        super().transform_content(contentnode)

        # Transform field lists.
        transformer = DocFieldTransformer(self)
        transformer.transform_all(contentnode)

        # Change line-breaks in top-level paragraphs into <br> nodes.
        for node in contentnode:
            if not isinstance(node, nodes.paragraph):
                continue

            line_broken = []
            for child in node.children:
                if isinstance(child, nodes.Text):
                    lines = child.astext().split('\n')
                    for i, line in enumerate(lines):
                        line_broken.append(nodes.Text(line))
                        if i < len(lines) - 1:
                            line_broken.append(
                                nodes.raw('', '<br />', format='html'))
                else:
                    line_broken.append(child)

            node.children = line_broken

    def run(self) -> list[Node]:
        # 親クラスの run() メソッドを呼び出す
        result = super().run()

        # signature parsing failed
        if not self.names:
            return result

        # Generate a target id for the function.
        encode_ = self.env.config.vb_encode_invalid_labels
        target_id = to_safe_label(self.names[0], encode_)

        # 表示用の関数名を作成
        function_name = str(self.names[0]).split('.')[-1] + '()'

        # 見出しを含むセクションノードを作成
        section_node = nodes.section()
        if self.env.config.vb_add_function_labels:
            # Add label to section node.
            if self.env.config.vb_add_docname_to_labels:
                delimiter = self.env.config.vb_docname_label_delimiter
                target_id = self.env.docname + delimiter + target_id
            section_node['names'].append(target_id)  # クロスリファレンス用
            section_node['ids'].append(target_id)    # HTML アンカー用
            section_node['label'] = target_id        # カスタム用途
        section_node += nodes.title(text=function_name)
        section_node += result

        # Add target to enable using implicit text (function_name)
        objects = self.env.domaindata['vb']['objects']
        objects[target_id] = (
            self.env.docname, target_id, 'function', function_name)

        # TODO:
        # これでローカルなターゲット id を作っているつもりだが上手く行かない。
        # グローバルなターゲット id はこれをしなくてもできている。
        # これの効果が不明。
        # ローカルなターゲット id は MyST でないと作られないのかも知れない。
        self.state.document.note_explicit_target(section_node)

        return [section_node]


class VBModule(Directive):
    '''For module directive (vb:module).

    Notes
    -----
    Not in use, just declared.
    '''
    # Class variables are module directive's specs.
    has_content = True  # for module summary or details.
    required_arguments = 1  # 1 arg is enough as whole declaration.
    option_spec = {}

    # def run(self):
    #     pass

    # def parse_arguments(self):
    #     pass

    # def parse_options(self):
    #     pass


class VBDomain(Domain):
    '''Domain for Visual Basic.
    '''
    name = 'vb'
    label = 'Visual Basic'

    object_types = {
        'function': ObjType('function', 'func', 'obj'),
    }
    directives = {
        'function': VBFunction,
        'module': VBModule,
    }
    roles = {
        'vbtype': XRefRole(innernodeclass=nodes.emphasis),
        'function': VBXRefRole(),
    }

    # autodoc で使われる情報を保持する辞書の、初期値
    initial_data = {
        # "functions": {},  # function name -> (docname, synopsis)
        # "classes": {},    # class name -> (docname, synopsis)
        # "modules": {},    # module name -> (docname, synopsis)
        "objects": {},    # object name -> (docname, objtype, signature)
    }

    def resolve_xref(
            self, env: BuildEnvironment, fromdocname: str, builder: Builder,
            typ: str, target: str, node: pending_xref, contnode: Element,
            ) -> Element | None:

        if target not in self.data['objects']:
            return None

        obj = self.data['objects'][target]
        if obj[2] != typ:  # obj[2] は登録時に指定したオブジェクトタイプ
            return None

        title = obj[3] if obj[2] == 'function' else target
        if 'refexplicit' in node.attributes and node.attributes['refexplicit']:
            child = contnode
        else:
            child = nodes.literal(text=title)

        result = make_refnode(
            builder, fromdocname, obj[0], obj[1], child, title)
        return result

    def resolve_any_xref(
            self, env, fromdocname, builder, target, node, contnode):
        results = []

        if target not in self.data['objects']:
            return results

        obj = self.data['objects'][target]
        title = obj[3] if obj[2] == 'function' else target
        if 'refexplicit' in node.attributes and node.attributes['refexplicit']:
            child = contnode
        else:
            child = nodes.literal(text=title)

        results.append(('', make_refnode(
            builder, fromdocname, obj[0], obj[1], child, title)))
        return results


def setup(app: Sphinx):
    '''Set up vb_domain feature.
    '''
    app.add_domain(VBDomain)

    # Config parameter to add function labels as reference targets.
    # This should be False if user enables sphinx.ext.autosectionlabel.
    app.add_config_value('vb_add_function_labels', True, 'env', bool)

    # Config parameter to encode labels that contain invalid characters.
    app.add_config_value('vb_encode_invalid_labels', True, 'env', bool)

    # Config parameter to add docname to function labels.
    # This should be True if there are multiple modules with same function
    # names.
    app.add_config_value('vb_add_docname_to_labels', False, 'env', bool)

    # Config parameter to set delimiter between module name and function name.
    app.add_config_value('vb_docname_label_delimiter', '-', 'env', str)



import hashlib
import re
from docutils.nodes import make_id


def to_safe_label(name: str, encode_: bool) -> str:
    '''Generate a valid label by encoding invalid characters with md5.'''
    invalid_symbols = re.compile(r'[!-/:-@\[-`{-~]')
    target_ptn = re.compile(r'[a-zA-Z0-9_\-]+')

    # Split name into parts.
    # e.g. 'Module1.Function1' -> ['Module1', 'Function1']
    name_parts = name.split('.')

    target_parts = []
    for name_part in name_parts:
        # ラベルに使えない記号をハイフンに置き換える。
        name_part = invalid_symbols.sub('-', name_part)

        if encode_:
            match = target_ptn.fullmatch(name_part)
            if match:
                target_part = name_part
            else:
                target_part = \
                    hashlib.md5(name_part.encode('utf-8')).hexdigest()[:8]
        else:
            target_part = name_part

        target_parts.append(make_id(target_part))

    return '.'.join(target_parts)



from sphinx.application import Sphinx

from .vb_autodoc import setup as setup_autodoc
from .vb_domain import setup as setup_domain

__version__ = '0.6.0'


def setup(app: Sphinx):
    '''Set up extension
    '''
    setup_autodoc(app)
    setup_domain(app)

    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from textwrap import dedent

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'vb_domain'
copyright = '2024, satamame.com'
author = 'satamame.com'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'myst_parser',
    # 'sphinx.ext.autosectionlabel',  # 自動でセクションラベルを付与する
    'sphinx_vb_domain',
]

templates_path = ['_templates']
exclude_patterns = []

language = 'ja'

# autosectionlabel_prefix_document = True  # セクションラベルにファイル名を付ける

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# -- Options for MyST-Parser -------------------------------------------------
myst_enable_extensions = [
    'fieldlist',
]
myst_heading_anchors = 3

# -- Options for sphinx_vb_domain --------------------------------------------

module_notes = {
    '__page__': '補足\n----\n\nページの補足説明',
    'Module1': '補足\n~~~~\n\nモジュールの補足説明',
    'Module1.procedureWithoutDocComment': dedent('''
        補足
        ^^^^

        - 関数の補足説明1
        - 関数の補足説明2
    '''),
}

vb_autodoc_paths = [
    ('../../macros/001', 'modules', 'モジュール', module_notes),
]

# vb_add_function_labels = False  # 関数のラベルを生成しない
# vb_encode_invalid_labels = False  # 無効なラベルをエンコードしない
# vb_add_docname_to_labels = True  # ラベルにファイル名を付ける
# vb_docname_label_delimiter = '__'  # ラベルのファイル名の区切り文字
# vb_autodoc_module_labels = True  # Autodoc でモジュールのラベルを生成する



import os
import re
import xml.etree.ElementTree as ET
from collections import namedtuple
from io import StringIO
from pathlib import Path
from typing import Iterator
from unicodedata import east_asian_width

from sphinx.application import Sphinx

from .utils import to_safe_label

# For config 'vb_autodoc_paths'. See `setup()` below.
AutodocPath = namedtuple('AutodocPath', ['src', 'rst', 'title', 'notes'])

# Regex pattern for function signature.
sig_ptn = re.compile(
    r'((Public|Private|Friend|Protected)\s+)?(Function|Sub)\s+([^\(\s]+)')


def xml_to_dict(xml_string) -> dict[str, str]:
    '''Convert document comment (xml) to dict.
    '''
    root = ET.fromstring(f'<root>{xml_string}</root>')
    result = {}
    for child in root:
        if child.tag == 'remarks' and 'remarks' in result:
            result['remarks'] += '\n' + child.text
        else:
            name = child.get('name', '')
            if name:
                result[f'{child.tag} {name}'] = child.text
            else:
                result[child.tag] = child.text
    return result


class DocComment:
    '''Object containing doc comment (xml) and/or func signature.
    '''
    def __init__(self, xml: str, sig: str):
        self.xml = xml
        self.sig = sig

    def get_param_type(self, param_name: str):
        '''Get paramter type from 'param As xx' part of signature.
        '''
        re_ptn = re.compile(param_name + r'\s+As\s+(\w+)')
        match = re_ptn.search(self.sig)
        if match and match.lastindex > 0:
            return match.group(1)
        else:
            return ''

    def get_return_type(self):
        '''Get return type from 'As xx' at end of signature.
        '''
        re_ptn = re.compile(r'As\s+(\w+)\s*$')
        match = re_ptn.search(self.sig)
        if match and match.lastindex > 0:
            return match.group(1)
        else:
            return ''

    def to_function_directive(self, module_name: str) -> str:
        indent = '   '
        content = f'.. vb:function:: {self.sig}\n'

        if module_name:
            content += f'{indent}:module: {module_name}\n'

        content += '\n'

        if not self.xml:
            return content

        xml_data = xml_to_dict(self.xml)

        if 'summary' in xml_data:
            summary_lines = xml_data['summary'].strip().split('\n')
            for line in summary_lines:
                content += f'{indent}{line.strip()}\n'
            content += '\n'

        content_length = len(content)

        for key in xml_data:
            if key.split()[0] in ('param', 'parameter', 'arg', 'argument'):
                param_name = key.split()[1]
                content += f'{indent}:{key}: {xml_data[key]}\n'

                param_type = self.get_param_type(param_name)
                if param_type:
                    content += f'{indent}:type {param_name}: {param_type}\n'
                continue

            if key in ('returns', 'return'):
                content += f'{indent}:{key}: {xml_data[key]}\n'
                return_type = self.get_return_type()
                if return_type:
                    content += f'{indent}:rtype: {return_type}\n'
                continue

            if key == 'rtype':
                content += f'{indent}:{key}: {xml_data[key]}\n'
                continue

        has_field_list = len(content) > content_length
        if has_field_list:
            content += '\n'

        if 'remarks' in xml_data:
            remark_lines = xml_data['remarks'].strip().split('\n')
            for line in remark_lines:
                content += f'{indent}{line}\n'
            content += '\n'

        return content

    def to_module_desc(self) -> str:
        content = ''
        xml_data = xml_to_dict(self.xml)

        if 'summary' in xml_data:
            summary_lines = xml_data['summary'].strip().split('\n')
            for line in summary_lines:
                content += f'{line}\n'
            content += '\n'

        if 'remarks' in xml_data:
            remark_lines = xml_data['remarks'].strip().split('\n')
            for line in remark_lines:
                content += f'{line}\n'
            content += '\n'

        return content

    def to_rest(self, module_name: str) -> str:
        if self.sig:
            return self.to_function_directive(module_name)
        else:
            return self.to_module_desc()

    @property
    def func_name(self) -> str:
        '''Extract the function name from the signature.'''
        if not self.sig:
            return ''
        match = sig_ptn.match(self.sig)
        if match:
            return match.group(4)  # Return the function name
        return ''


def extract_doccomments(f: StringIO) -> Iterator[DocComment]:
    '''Generator of Document Comments from a text stream
    '''
    xml = ''
    eof = False

    while not eof:
        line = f.readline()
        if not line:
            eof = True
        line = line.strip()

        # Extract document comment (maybe empty) and function signature.
        if sig_ptn.match(line):
            yield DocComment(xml, line)
            xml = ''
            continue

        # Keep part of document comment in xml.
        if line.startswith("'''"):
            if xml:
                xml += '\n'
            xml += line[3:].strip()
            continue

        # Neither function signature nor xml: yield document comment if kept.
        if xml:
            yield DocComment(xml, '')
            xml = ''


def headline_len(title: str) -> int:
    def char_width(char: str) -> int:
        return 2 if east_asian_width(char) in ('F', 'W') else 1

    return sum(char_width(char) for char in title)


def sanitize_note(note: str) -> str:
    '''Remove or demote headings in the note content.'''
    # Replace reST heading patterns with plain text
    sanitized_note = re.sub(
        r'^(.*)\n[-=~^`]+\n', r'**\1**\n', note, flags=re.MULTILINE)
    return sanitized_note


def generate_module_content(
        src_file: Path, module_name: str, autodoc_path: AutodocPath,
        app: Sphinx) -> str:
    '''Generate reST content per module

    Parameters
    ----------
    src_file : Path
        Path to VB module source file.
    module_name : str
        Module name to show in document.
    autodoc_path : AutodocPath
        AutodocPath object to be handled.
    app : Sphinx
        Sphinx application object.

    Returns
    -------
    module_content : str
        Document content in reStructuredText for the module.
    '''
    # Module headline (level2)
    content = f"\n{module_name}\n{'-' * headline_len(module_name)}\n\n"

    # Add label to module section if enabled.
    if app.config.vb_autodoc_module_labels:

        # Generate a target id for the function.
        encode_ = app.config.vb_encode_invalid_labels
        target_id = to_safe_label(module_name, encode_)

        if app.config.vb_add_docname_to_labels:
            delimiter = app.config.vb_docname_label_delimiter
            label = autodoc_path.rst + delimiter + target_id
        else:
            label = target_id
        content = f".. _{label}:\n\n" + content

    # Add note to module block using notes.
    module_note = autodoc_path.notes.get(module_name)
    if module_note:
        content += f"{sanitize_note(module_note)}\n\n"

    with open(src_file, 'r', encoding='utf-8') as f:
        for doccomment in extract_doccomments(f):
            content += doccomment.to_rest(module_name)
            func_note = autodoc_path.notes.get(
                f'{module_name}.{doccomment.func_name}')
            if func_note:
                content += f"{sanitize_note(func_note)}\n\n"

    return content


def generate_rst_files(app: Sphinx):
    '''Create/overwrite *.rst files based on VB source directory.

    This is called just after the builder is inited.
    '''
    if not app.config.vb_autodoc:
        return

    for path_info in app.config.vb_autodoc_paths:
        if len(path_info) < 3:
            raise ValueError('vb_autodoc_paths must have at least 3 elements.')
        if len(path_info) < 4:
            path_info = (*path_info, {})

        autodoc_path = AutodocPath(*path_info)
        title = autodoc_path.title
        rst_content = f"{title}\n{'=' * headline_len(title)}\n\n"

        # Add note to rst_content using notes.
        page_note = autodoc_path.notes.get('__page__')
        if page_note:
            rst_content += f"{sanitize_note(page_note)}\n\n"

        src_dir = Path(app.confdir) / autodoc_path.src
        for vb_file in os.listdir(src_dir):
            if vb_file.endswith(('.bas', '.vb', '.vbs')):
                module_name = os.path.splitext(os.path.basename(vb_file))[0]
                src_file = src_dir / vb_file
                rst_content += generate_module_content(
                    src_file, module_name, autodoc_path, app)

        dest_file = Path(app.srcdir) / (autodoc_path.rst + '.rst')
        with open(dest_file, 'w', encoding='utf-8') as f:
            f.write(rst_content)


def setup(app: Sphinx):
    '''Set up vb_autodoc feature.
    '''
    # Config parameter to enable autodoc.
    # > sphinx-build html docs/source docs/build -D vb_autodoc=1
    app.add_config_value('vb_autodoc', False, 'env', bool)

    # Config parameter to set source dirs, rst names, titles, and callback.
    # e.g. [('../../macros', 'modules', 'Modules', notes)]
    #    makes 'modules.rst' from '../../macros/*.bas' files.
    # '../../macros' should be relative from Sphinx conf dir.
    # 'modules' is treated as 'modules.rst' relative from Sphinx src dir.
    # 'Modules' will be the title of 'modules.rst' page.
    # notes is a dict to map notes to targets where to add.
    # e.g. {'__page__': 'page note', 'Module Name': 'module note'}
    app.add_config_value('vb_autodoc_paths', [], 'env', list[AutodocPath])

    # Config parameter to add module labels as reference targets.
    app.add_config_value('vb_autodoc_module_labels', False, 'env', bool)

    # Add process just after the builder is inited.
    app.connect('builder-inited', generate_rst_files)



from sphinx_vb_domain.vb_domain import VBDomain


# def test_handle_signature():
#     vb_function = VBDomain.VBFunction()
#     result = vb_function.handle_signature(
#         "Public Function MyFunction(x As Integer) As String")
#     assert result['access_modifier'] == 'Public'
#     assert result['func_type'] == 'Function'
#     assert result['func_name'] == 'MyFunction'
#     assert result['args'] == 'x As Integer'
#     assert result['return_type'] == 'String'


def test_dummy():
    assert True
